# -*- coding: utf-8 -*-
#
# Copyright (c) 2018~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

import sys
from typing import List, Tuple
from inspect import signature, Parameter

def inject_by(func, required: List[Tuple[str, object]], optional: List[Tuple[str, object, object]]):
    '''
    `required`: `(name, key)`
    `optional`: `(name, key, default)`
    '''
    def new_func(ioc):
        kwargs = {}
        for name, key in required:
            kwargs[name] = ioc[key]
        for name, key, default in optional:
            kwargs[name] = ioc.get(key, default)
        return func(**kwargs)
    return new_func

def inject_by_name(func):
    '''
    wrap the func and auto inject by parameter name.

    var keyword parameter and var positional parameter will not be inject.

    return the new func with signature: `(ioc) => any`
    '''
    sign = signature(func)
    params = [p for p in sign.parameters.values() if p.kind in (
        Parameter.POSITIONAL_OR_KEYWORD,
        Parameter.KEYWORD_ONLY
    )]
    required: List[Tuple[str, str]] = []
    optional: List[Tuple[str, str, object]] = []
    for param in params:
        if param.default is Parameter.empty:
            required.append((param.name, param.name))
        else:
            optional.append((param.name, param.name, param.default))
    return inject_by(func, required, optional)

def inject_by_anno(func):
    '''
    wrap the func and auto inject by parameter annotation.

    var keyword parameter and var positional parameter will not be inject.

    return the new func with signature: `(ioc) => any`
    '''
    sign = signature(func)
    params = [p for p in sign.parameters.values() if p.kind in (
        Parameter.POSITIONAL_OR_KEYWORD,
        Parameter.KEYWORD_ONLY
    )]
    required: List[Tuple[str, object]] = []
    optional: List[Tuple[str, object, object]] = []
    for param in params:
        if param.annotation is Parameter.empty:
            if param.default is Parameter.empty:
                raise ValueError(f'annotation of args {param.name} is empty.')
            else:
                # use `object()` for ensure you never get the value.
                optional.append((param.name, object(), param.default))
        else:
            if param.default is Parameter.empty:
                required.append((param.name, param.annotation))
            else:
                optional.append((param.name, param.annotation, param.default))
    return inject_by(func, required, optional)

def auto_enter(func):
    '''
    auto enter the context manager when it created.

    the signature of func should be `(ioc) => any`.
    '''
    def new_func(ioc):
        item = func(ioc)
        ioc.enter(item)
        return item
    return new_func

def dispose_at_exit(provider):
    '''
    register `provider.__exit__()` into `atexit` module.

    return the `provider` itself.
    '''
    import atexit
    @atexit.register
    def provider_dispose_at_exit():
        provider.__exit__(*sys.exc_info())
    return provider

def make_group(container, group_key):
    '''
    add a new group into `container` by `group_key`,
    return a decorator function for add next group item key.

    '''
    group_keys = []
    container.register_group(group_key, group_keys)
    def decorator(next_group_key):
        group_keys.append(next_group_key)
        return next_group_key
    return decorator

# keep old func names:

auto_inject = inject_by_name



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

from abc import abstractmethod
from enum import Enum
from inspect import signature, Parameter
from typing import Any

from .symbols import Symbols


class LifeTime(Enum):
    transient = 0
    scoped = 1
    singleton = 2


class IServiceInfo:
    @abstractmethod
    def get(self, provider) -> Any:
        raise NotImplementedError


class ServiceInfo(IServiceInfo):
    def __init__(self, key, factory, lifetime):

        sign = signature(factory)
        if not sign.parameters:
            self._factory = lambda _: factory()
        elif len(sign.parameters) == 1:
            arg_0 = list(sign.parameters.values())[0]
            if arg_0.kind != Parameter.POSITIONAL_OR_KEYWORD:
                raise TypeError('1st parameter of factory must be a positional parameter.')
            self._factory = factory
        else:
            raise TypeError('factory has too many parameters.')

        self._key = key
        self._lifetime = lifetime
        self._cache_value = None

    def get(self, provider):
        if self._lifetime is LifeTime.transient:
            return self._factory(provider)

        if self._lifetime is LifeTime.scoped:
            cache = provider[Symbols.cache]
            if self not in cache:
                cache[self] = self._factory(provider)
            return cache[self]

        if self._lifetime is LifeTime.singleton:
            if self._cache_value is None:
                provider = provider[Symbols.provider_root]
                self._cache_value = (self._factory(provider), )
            return self._cache_value[0]

        raise NotImplementedError(f'what is {self._lifetime}?')


class ProviderServiceInfo(IServiceInfo):
    def get(self, provider):
        return provider


class ValueServiceInfo(IServiceInfo):
    def __init__(self, value):
        self._value = value

    def get(self, provider):
        return self._value


class GroupedServiceInfo(IServiceInfo):
    def __init__(self, keys: list):
        self._keys = keys

    def get(self, provider):
        return tuple(provider[k] for k in self._keys)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018~2999 - Cologler <skyoflw@gmail.com>
# ----------
# a global ioc
# ----------

from .ioc import ServiceProvider
from .utils import inject_by_name, dispose_at_exit

ioc = ServiceProvider()
dispose_at_exit(ioc)

def ioc_singleton(func):
    ioc.register_singleton(func.__name__, func)
    return func

def ioc_scoped(func):
    ioc.register_scoped(func.__name__, func)
    return func

def ioc_transient(func):
    ioc.register_transient(func.__name__, func)
    return func

def ioc_singleton_cls(wrap=inject_by_name):
    def wrapper(cls: type):
        wraped_cls = wrap(cls) if wrap else cls
        ioc.register_singleton(cls, wraped_cls)
        ioc.register_singleton(cls.__name__, lambda x: x[cls])
        return cls
    return wrapper

def ioc_scoped_cls(wrap=inject_by_name):
    def wrapper(cls: type):
        wraped_cls = wrap(cls) if wrap else cls
        ioc.register_scoped(cls, wraped_cls)
        ioc.register_scoped(cls.__name__, lambda x: x[cls])
        return cls
    return wrapper

def ioc_transient_cls(wrap=inject_by_name):
    def wrapper(cls: type):
        wraped_cls = wrap(cls) if wrap else cls
        ioc.register_transient(cls, wraped_cls)
        ioc.register_transient(cls.__name__, lambda x: x[cls])
        return cls
    return wrapper

def ioc_bind(new_key):
    '''
    bind with new key.
    '''
    def binding(cls):
        name = cls.__name__
        ioc.register_transient(new_key, lambda x: x[name])
        return cls
    return binding



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

from abc import abstractmethod
from typing import List

from .err import ServiceNotFoundError
from .ioc_service_info import ValueServiceInfo

class IServiceInfoResolver:
    def get(self, provider, key):
        '''
        get the `IServiceInfo` from resolver.
        '''
        raise ServiceNotFoundError(key)

    def __add__(self, other):
        new_resolver = ServiceInfoChainResolver()
        new_resolver.chain.append(self)
        new_resolver.append(other)
        return new_resolver

    def cache(self):
        '''
        return a `CacheServiceInfoResolver`.
        '''
        return CacheServiceInfoResolver(self)


class ServiceInfoChainResolver(IServiceInfoResolver):
    '''
    the chain resolver
    '''
    def __init__(self, *resolvers):
        self.chain: List[IServiceInfoResolver] = list(resolvers)

    def get(self, provider, key):
        for resolver in self.chain:
            try:
                return resolver.get(provider, key)
            except ServiceNotFoundError:
                pass
        return super().get(provider, key)

    def append(self, other):
        if isinstance(other, ServiceInfoChainResolver):
            self.chain.extend(other.chain)
        else:
            self.chain.append(other)

    def __add__(self, other):
        new_resolver = ServiceInfoChainResolver()
        new_resolver.chain.extend(self.chain)
        new_resolver.append(other)
        return new_resolver


class CacheServiceInfoResolver(IServiceInfoResolver):
    '''
    NOTE: if a `IServiceInfo` is affect by `provider`, you should not cache it.
    `CacheServiceInfoResolver` only cache by the `key` and ignore the `provider` arguments.
    '''
    def __init__(self, base_resolver: IServiceInfoResolver):
        super().__init__()
        self._base_resolver = base_resolver
        self._cache = {}

    def get(self, provider, key):
        try:
            return self._cache[key]
        except KeyError:
            pass
        service_info = self._base_resolver.get(provider, key)
        self._cache[key] = service_info
        return service_info

    def cache(self):
        return self


class ImportServiceInfoResolver(IServiceInfoResolver):
    def get(self, provider, key):
        import importlib
        if isinstance(key, str):
            try:
                module = importlib.import_module(key)
                return ValueServiceInfo(module)
            except TypeError:
                pass
            except ModuleNotFoundError:
                pass
        return super().get(provider, key)


class TypesServiceInfoResolver(IServiceInfoResolver):
    def get(self, provider, key):
        if isinstance(key, type):
            from .ioc_service_info import ServiceInfo, LifeTime
            from .utils import inject_by_name
            ctor = inject_by_name(key)
            return ServiceInfo(key, ctor, LifeTime.transient)
        return super().get(provider, key)
